# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   06-Sep-2006

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

import os
import tkinter as tk

import icons
import tkExtra
import tkFlair
import Project

_CHILD = "."

#===============================================================================
# Run Listbox
#===============================================================================
class RunListbox(tkExtra.MultiListbox):
	def __init__(self, master=None, edit=False, **kw):
		super().__init__(master,
				(("+",    1, None),
				 ('Run', 20, None),
				 ('Spawn',3, None)),
				 **kw)
		self.sortAssist = None

		# Bind events
		self.bindList("<space>",		self.toggle)
		self.bindList("<<Rename>>",		self.renameInPlace)
		self.bindList("<ButtonRelease-1>",	self.buttonRelease)
		self.bindLeftRight()

		self.listbox(1).bind("<Return>",	self.renameInPlace)
		self.listbox(1).bind("<KP_Enter>",	self.renameInPlace)
		self.listbox(1).bind("<Double-1>",	self.renameInPlace)
		self.listbox(1).bind("<Control-Key-e>",	self.renameInPlace)

		self.listbox(2).bind("<<Rename>>",	self.spawnInPlace)
		self.listbox(2).bind("<Return>",	self.spawnInPlace)
		self.listbox(2).bind("<KP_Enter>",	self.spawnInPlace)
		self.listbox(2).bind("<Button-1>",	self.spawnButton1)

		self.setPopupMenu([
			#("Expand",     0, self.expand,   icons.get("+")),
			#("Collapse",   0, self.collapse, icons.get("-")),
			#None,
			#("Add",        0, self.add,           icons.get("add")),
			#("Remove",     0, self.remove,        icons.get("x")),
			#("Clone",      0, self.clone,         icons.get("clone")) ])
			("Rename",     0, self.renameInPlace, icons.get("rename")) ])

		# Remember selected runs
		self.edit     = edit
		self.selected = []
		self.expanded = []
		self.flair    = None
		self.project  = None
		self.lastrelease = -1

	# ---------------------------------------------------------------------
	def set(self, flair):
		self.flair = flair
		self.project = flair.project
		self.fill()

	# ----------------------------------------------------------------------
	# fill run list
	# ----------------------------------------------------------------------
	def fill(self):
		active = self.index(tk.ACTIVE)
		self.delete(0, tk.END)
		for run in self.project.runs:
			# ignore child
			if run("parent"): continue
			name = run.inputName2()
			if run.family:
				if name in self.expanded:
					exp = "-"
				else:
					exp = "+"
				self.insert(tk.END, (exp, name, str(len(run.family))))
				try:
					self.itemconfig(tk.END,
						foreground=Project.RunInfo._statusColor[run("status")])
				except KeyError:
					pass
				if name in self.selected: self.selection_set(tk.END)
				if exp == "-":
					for child in run.family:
						childRun = self.project.getRunByName(child, run.name)
						self.insert(tk.END, (_CHILD, f"  {child}",""))
						try:
							self.itemconfig(tk.END,
								foreground=Project.RunInfo._statusColor[childRun("status")])
						except KeyError:
							pass
						if child in self.selected:
							self.selection_set(tk.END)
			else:
				self.insert(tk.END, ("",name,""))
				try:
					self.itemconfig(tk.END,
						foreground=Project.RunInfo._statusColor[run("status")])
				except KeyError:
					pass
				if name in self.selected: self.selection_set(tk.END)

		if not self.curselection():
			self.selection_set(0)
			active = 0
		self.activate(active)
		self.see(active)

	# ----------------------------------------------------------------------
	# Update run colors
	# ----------------------------------------------------------------------
	def updateColors(self):
		for i in range(self.size()):
			run = self.getRun(i)
			if run is not None:
				try:
					self.itemconfig(i,
						foreground=Project.RunInfo._statusColor[run("status")])
				except KeyError:
					pass

	# ----------------------------------------------------------------------
	# Expand selected items
	# ----------------------------------------------------------------------
	def expand(self):
		sel = list(map(int, self.curselection()))
		if not sel: return

		sel.reverse()
		for i in sel:
			if self.listbox(0).get(i) != "+": continue
			self.expanded.append(self.listbox(1).get(i))

		self.updateSelection()
		self.fill()
		self.activate(sel[-1])

	# ----------------------------------------------------------------------
	# Collapse selected items
	# ----------------------------------------------------------------------
	def collapse(self):
		sel = list(map(int, self.curselection()))
		if not sel: return
		sel.reverse()

		for i in sel:
			if self.listbox(0).get(i) != "-": continue
			try: self.expanded.remove(self.listbox(1).get(i))
			except ValueError: pass

		self.updateSelection()
		self.fill()
		self.activate(sel[-1])

	# ----------------------------------------------------------------------
	def toggle(self, event=None):
		sel = list(map(int, self.curselection()))
		if not sel: return
		sel.reverse()

		expand = None
		for i in sel:
			item = self.listbox(0).get(i)
			if item not in ("+","-"): continue
			name = self.listbox(1).get(i)

			if expand is None:
				if item not in ("+","-"): break
				expand = item == "+"

			if expand:
				self.expanded.append(name)
			else:
				try: self.expanded.remove(name)
				except ValueError: pass

		self.updateSelection()
		self.fill()
		self.activate(sel[-1])
		return "break"

	# ----------------------------------------------------------------------
	def buttonRelease(self, event):
		listbox = event.widget
		act = listbox.nearest(event.y)
		self.activate(act)
		lid = self.listboxIndex(listbox)

		if lid==0:
			if listbox.get(act) in ("+","-"): self.toggle()
		elif lid==2:
			# Change the number of spawned jobs
			if self.listbox(0).get(act) != _CHILD:
				self.spawnInPlace()

	# ----------------------------------------------------------------------
	# Update/remember selection
	# ----------------------------------------------------------------------
	def updateSelection(self):
		del self.selected[:]
		for i in self.curselection():
			self.selected.append(self.listbox(1).get(i).strip())

	# ----------------------------------------------------------------------
	def getName(self, pos):
		return self.listbox(1).get(pos).strip()

	# ----------------------------------------------------------------------
	# Return run from position
	# ----------------------------------------------------------------------
	def getRun(self, pos):
		idx = int(self.index(pos))
		if idx==0: return self.project.runs[0]
		name = self.listbox(1).get(idx).strip()
		if self.listbox(0).get(idx) == _CHILD:
			# Child find parent
			for i in range(idx-1,0,-1):
				if self.listbox(0).get(i) in ("+","-"):
					parent = self.listbox(1).get(i).strip()
					break
			else:
				parent = self.project.runs[0].name
		else:
			parent = ""
		return self.project.getRunByName(name, parent)

	# ----------------------------------------------------------------------
	# Return run list selection
	# ----------------------------------------------------------------------
	def runSelection(self):
		runs = []
		for i in self.curselection():
			run = self.getRun(i)
			if run is None: continue
			if run.family and run.name not in self.expanded:
				runs.append(run)
				for child in run.family:
					runs.append(self.project.getRunByName(child,
						run.name))
			else:
				runs.append(run)
		return runs

	# ----------------------------------------------------------------------
	def spawnButton1(self, event):
		act = self.listbox(2).nearest(event.y)
		if self.listbox(2).selection_includes(act):
			self.activate(act)
			return "break"

	# ----------------------------------------------------------------------
	# Hitting F2 rename in place
	# ----------------------------------------------------------------------
	def renameInPlace(self, event=None):
		if not self.edit: return

		sel = list(map(int,self.curselection()))
		if len(sel) != 1 or sel[0]==0: return "break"
		pos = sel[0]
		self.activate(pos)

		item = self.listbox(0).get(pos)
		if item==_CHILD: return "break"		# Ignore child

		name = self.listbox(1).get(pos)
		run = self.project.getRunByName(name.strip())
		if run is None or run("parent"): return "break"

		# Set the correct name
		self.listbox(1).set(pos, name.strip())
		self.listbox(1).focus_set()
		edit = tkExtra.InPlaceFile(self.listbox(1),
				title=f"Rename run '{run.name}' to",
				filetypes=[("Fluka Input files","*.inp"),
					("All","*")])

		if edit.value is not None and edit.value != name.strip():
			f = self.project.relpath(edit.value)
			n,e = os.path.splitext(f)
			self.renameRun(run, n)
			self.fill()
			self.flair.modified()
		else:
			self.listbox(1).set(pos, name)
		self.flair.refresh("run")
		return "break"

	# ----------------------------------------------------------------------
	# Rename a run
	# @param run	to rename
	# @param name	new name
	# ----------------------------------------------------------------------
	def renameRun(self, run, name):
		# Check if run is there
		newname = self.project.findUniqueRunName(name)
		if newname != name:
			self.flair.notify("Run already exists",
				f"Run '{name}' already exists. Renamed to '{newname}'",
				tkFlair.NOTIFY_WARNING)
			name = newname

		self.updateSelection()

		try: self.selected.remove(run.name)
		except ValueError: pass
		self.selected.append(name)

		try:
			self.expanded.remove(run.name)
			self.expanded.append(name)
		except ValueError:
			pass

		self.project.renameRun(run,name)
		self.flair.refresh("run")

	# ----------------------------------------------------------------------
	# Number of spawned runs has been edited
	# ----------------------------------------------------------------------
	def spawnInPlace(self, event=None):
		if not self.edit: return

		#self.get()
		# protect against accidental delete
		#if self.flair.event is not None:
		#	if self.focus_get() is not self.runList: return

		sel = list(map(int, self.curselection()))
		if not sel: return
		active = self.index(tk.ACTIVE)
		if active not in sel:
			self.activate(sel[0])
		self.listbox(2).focus_set()

		try:    oldSpawn = int(self.listbox(2).get(sel[0]))
		except: oldSpawn = 0
		edit = tkExtra.InPlaceEdit(self.listbox(2))

		if edit.value is not None:
			try:
				spawn = int(edit.value)
			except:
				spawn = 0
		else:
			return

		if spawn == oldSpawn: return

		sel.reverse()
		for pos in sel:
			run = self.getRun(pos)
			if run("parent"): continue
			index = self.project.getRunIndex(run)

			# Remove additional runs
			for name in run.family[spawn:]:
				child = self.project.getRunByName(name, run.name)
				if child is not None:
					self.project.delRun(child)

			# Fill runs
			for i in range(len(run.family),spawn):
				name = self.project.getSpawnRunName(run,i+1)
				if self.project.getRunByName(name, run.name) is not None:
					continue

				child = Project.RunInfo(self.project, name)
				child["parent"] = run.name
				self.project.setRunRandomSeed(child, run, i+1)
				self.project.addRun(child, index+i+1)
				run.family.append(child.name)

		self.flair.modified()
		self.updateSelection()
		self.fill()
