# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

from log import error

CONVERSION_CLASSES = {}

#-------------------------------------------------------------------------------
# Register conversion class
#-------------------------------------------------------------------------------
def register(src, dst, class_):
	global CONVERSION_CLASSES
	key = "%s:%s"%(src,dst)
	CONVERSION_CLASSES[key] = class_

#-------------------------------------------------------------------------------
def conversionClass(src, dst):
	return CONVERSION_CLASSES["%s:%s"%(src,dst)]

#	# ----------------------------------------------------------------------
#	# Convert input file to mode
#	# ----------------------------------------------------------------------
#	def convertTo(self, mode):
#		if mode==self.mode or mode=="":
#			return self.clone()
#		if mode=="moira":
#			return self.convert2Moira()
#		elif mode=="fluka":
#			return self.convert2Fluka()
#		elif mode=="mcnp":
#			return self.convert2Mcnp()
#		else:
#			return None
#
#	# ----------------------------------------------------------------------
#	def convert2Moira(self):
#		return None
#
#	# ----------------------------------------------------------------------
#	def convert2Fluka(self):
#		return None
#
#	# ----------------------------------------------------------------------
#	def convert2Mcnp(self):
#		return None
#
#
#===============================================================================
# Basic abstract class for conversion from one format to another
#===============================================================================
class Convert:
	src = ""
	dst = ""

	#-----------------------------------------------------------------------
	def __init__(self, inp, defines=None):
		self.input   = inp
		self.defines = defines
		self.output  = None

	#-----------------------------------------------------------------------
	def convert(self):
		return self.output

	#-----------------------------------------------------------------------
	def notImplemented(self, card, what, msg="", notimpl="not implemented"):
		if isinstance(what,int):
			error(card.tag,
				f"what[{what}]={card.what(what)}",
				notimpl,msg)
		else:
			error(card.tag,what,notimpl,msg)

	#-----------------------------------------------------------------------
	def add(self, card):
		self.output.add(card)

	#-----------------------------------------------------------------------
	def addClone(self, card, disable=False):
		clone = card.clone()
		if disable: clone.setEnable(False)
		self.add(card)
