# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Nov-2018

import os
import re

#===============================================================================
# A container class to keep the programs
#===============================================================================
class Config:
	SECTION   = ""		# Configuration section name
	directory = ""		# Executable installation directory
#	var       = ""		# environment variable
	program   = ""		# Main application program
	version   = ""		# Default program version
	released  = ""
	config    = None	# Configuration file

	CORE_PAT  = re.compile(r"^core\.?\d*$")
	programs  = {}

	#-------------------------------------------------------------------------------
	@staticmethod
	def getInt(section, name, default):
		try:
			return Config.config.getint(section, name)
		except:
			Config.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getFloat(section, name, default):
		try:
			return Config.config.getfloat(section, name)
		except:
			Config.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getStr(section, name, default):
		try:
			return Config.config.get(section, name)
		except:
			Config.config.set(section, name, str(default))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def getBool(section, name, default):
		try:
			return Config.config.getboolean(section, name)
		except:
			Config.config.set(section, name, str(bool(default)))
			return default

	#-------------------------------------------------------------------------------
	@staticmethod
	def set(section, name, value):
		Config.config.set(section, name, str(value))
	setStr   = set
	setInt   = set
	setFloat = set

	#-------------------------------------------------------------------------------
	@staticmethod
	def setBool(section, name, value):
		Config.config.set(section, name, str(bool(value)))

	#=======================================================================
	# Configuration related methods
	#=======================================================================
	@staticmethod
	def findVersion(directory=None):
		pass

	#-----------------------------------------------------------------------
	@staticmethod
	def command(prg):
		# Try flair directory
		cmd = os.path.join(os.path.dirname(os.path.dirname(__file__)), prg)
		if os.path.isfile(cmd): return cmd
		return prg

	#-----------------------------------------------------------------------
	def executable(self):
		return os.path.join(self.directory, self.program)
