# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import Evaluation
from log import say

#===============================================================================
# Particle class
#===============================================================================
class Particle:
	# Particles dictionary
	_db           = {}
	list          = []
	hadron        = []
	lepton        = []
	beam          = []
	listAll       = []
	signedList    = []
	signedListAll = []
	particles     = []

	# ----------------------------------------------------------------------
	def __init__(self, idx, name, mass=0.0, comment=None, pdg=0):
		self.id      = idx
		self.name    = name
		self.mass    = mass
		self.comment = comment
		self.pdg     = pdg

	# ----------------------------------------------------------------------
	# add particle
	# ----------------------------------------------------------------------
	@staticmethod
	def add(idx, name, mass=0.0, comment=None, pdg=0):
		p = Particle(idx, name, mass, comment, pdg)
		Particle._db[idx]  = p
		Particle._db[name] = p

		# Populate also _baseLocalDict
		if mass>0.0:
			n = name
			if name[-1] == "-":
				n = n[:-1]+"n"
			elif name[-1] == "+":
				n = n[:-1]+"p"
			if n.find("-")>=0:
				n = n.replace("-","")
			Evaluation._globalDict["m%s"%(n)] = mass
		return p

	# ----------------------------------------------------------------------
	def __getitem__(self, idx):
		return Particle.particles[idx]

	# ----------------------------------------------------------------------
	@staticmethod
	def get(idx):
		return Particle._db.get(idx)

	# ----------------------------------------------------------------------
	# create lists
	# ----------------------------------------------------------------------
	@staticmethod
	def makeLists():
		if Particle.list: return
		lastpar = Particle.convert("@LASTPAR",False)
		# All particles
		Particle.list.append("")
		Particle.hadron.append("")
		Particle.lepton.append("")
		Particle.listAll.append("")
#		for p in Particle._db:
#			if isinstance(p, str):
		# Go through in ordered by id list
		for i in sorted([Particle._db[x].id
				for x in Particle._db
					if isinstance(x,str)]):
			part = Particle._db[i]
			if part.name != "@LASTPAR":
				Particle.listAll.append(part.name)
			if part.id < lastpar:
				Particle.list.append(part.name)
			if abs(part.pdg)<100:
				Particle.lepton.append(part.name)
			else:
				Particle.hadron.append(part.name)

		# Beam particles
		Particle.beam = Particle.list[:]
		Particle.beam.append("ISOTOPE")

		Particle.hadron.append("4-HELLIUM")
		Particle.hadron.append("3-HELLIUM")
		Particle.hadron.append("TRITON")
		Particle.hadron.append("DEUTERON")
		Particle.hadron.append("HEAVYION")

		# Sort lists
		Particle.beam.sort()
		Particle.hadron.sort()
		Particle.lepton.sort()

		# Only real particles + generalized ones
		Particle.particles = Particle.list[:]
		for p in range(201, 228):
			if p in (208, 211, 219, 220, 221, 222): continue
			Particle.particles.append(Particle._db[p].name)

		# Sort
		Particle.list.append("@LASTPAR")
		Particle.listAll.append("@LASTPAR")

		# Create signed list
		Particle.signedList += Particle.list
		for i in range(len(Particle.signedList)-1, 0, -1):
			p = Particle.signedList[i]
			if p and p[0] != '@':
				Particle.signedList.insert(i+1, '-'+p)

		Particle.signedListAll += Particle.listAll
		for i in range(len(Particle.signedListAll)-1, 0, -1):
			p = Particle.signedListAll[i]
			if p and p[0] != '@':
				Particle.signedListAll.insert(i+1, '-'+p)

	# ----------------------------------------------------------------------
	# Convert to integer or string
	# ----------------------------------------------------------------------
	@staticmethod
	def convert(pid, tonames=True):
		"""Convert particle from name to number
		pid particle type
		tonames True/False"""

		try:
			idx = int(pid)
			isInt = True
		except:
			if pid == "":
				idx = 0
				isInt = True
			else:
				isInt = False

		if tonames:
			if isInt:
				if idx<=-100 or idx>1000: return str(pid)
				#if 988 <= idx <= 1012:
				#	# Only for DISCARD card
				#	idx = -abs(1000 - idx)
				try:
					return Particle._db[idx].name
				except:
					say(f"Particle {pid} not found!")
					return "UNKNOWN"
			else:
				return pid
		else:
			if isInt:
				return idx
			try:
				return Particle._db[pid].id
			except:
				raise Exception(f"Particle {pid} not found!")

	# ----------------------------------------------------------------------
	# scan input for particles
	# ----------------------------------------------------------------------
	@staticmethod
	def scan(input_):
		# FIXME for the moment return everything
		return Particle.listAll
