# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   24-Oct-2018

__author__ = "Vasilis Vlachoudis"
__email__  = "Vasilis.Vlachoudis@cern.ch"

__all__ = [	"Moira",
		"MoiraCard",
		"MoiraCardInfo",
		"MoiraConfig",
		"MoiraLayout",
		"Gdml",
		"Gdml2Moira",
		"Gdml2Fluka"]

from itertools import count

from .input  import Moira
from .card   import MoiraCard, MoiraCardInfo
from .gdml   import Gdml, Gdml2Moira, Gdml2Fluka
from .layout import MoiraLayout
from .config import MoiraConfig

DATABASE     = "db/moira.ini"
G4DATABASE   = "db/geant4.ini"
NISTDATABASE = "db/nist.ini"

#-------------------------------------------------------------------------------
# Initialize classes:
#	MoiraCardInfo
#	MoiraParticle
#-------------------------------------------------------------------------------
def init(filename=None, g4name=None, nistname=None):
	import configparser
	from deck import Card, dbpath, readCardini

	Card.register_class(Moira.MODE, MoiraCard)

	# Load card database
	if filename is None:
		filename = dbpath(DATABASE)
	if g4name is None:
		g4name   = dbpath(G4DATABASE)
	if nistname is None:
		nistname = dbpath(NISTDATABASE)

	# read card database and prepare CardInfo classes
	cardini = configparser.RawConfigParser()
	cardini.optionxform = str	# make case sensitive
	if not cardini.read((nistname, g4name, filename)): return False

	# parse cards
	readCardini(cardini, Moira.MODE, MoiraCardInfo)

	# Load nist material database
	#if nistname is None:
	#	nistname = os.path.join(os.path.dirname(__file__), "nist.ini")

	#cardini = configparser.RawConfigParser()
	#if not cardini.read(nistname): return False

	# Read default materials and create fake MATERIAL card's
	section = "@materials"
	for i in count(1):
		n = f"name.{i}"
		try: name = cardini.get(section, n)
		except: break
		rho = cardini.getfloat(section, f"rho.{i}")
		#Z  = cardini.getfloat(section, f"Z.{i}")
		#I  = cardini.get(     section, f"I.{i}")
		card = MoiraCard("MIXT", [name, rho, 0])
		Moira.defaultMaterials[name] = card

	# Read FLUKA to G4 conversion for materials
	section = "@fluka-materials"
	for name, value in cardini.items(section):
		Moira.addFlukaMat2Moira(name, value)

	# ... and particles
	section = "@fluka-particles"
	for name, value in cardini.items(section):
		Moira.addFlukaParticle2Moira(name, value)

	del cardini
	return True

#===============================================================================
# Load only once
#init()
