# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Nov-2018

import os
import re

from log import error
from deck import Config

#===============================================================================
# A container class to keep the programs
# It should contain only static and class methods
#===============================================================================
class MoiraConfig(Config):
	SECTION   = "Moira"	# Configuration section name

	DEFAULT_EXE = "moira"

	CYCLE     = "moira_"	# prefix of temporary cycle directory

	directory = ""		# Executable installation directory
	var       = ""		# environment variable
	program   = DEFAULT_EXE	# Main application program
	version   = "Unknown"
	released  = ""

	CYCLEDIR_PAT  = re.compile(r"^%s(\d+)$"%(CYCLE))
	CYCLEDIR2_PAT = re.compile(r"^(.*/?)%s(\d+)$"%(CYCLE))

	#-----------------------------------------------------------------------
	@staticmethod
	def checkDir():
		try:
			os.stat(MoiraConfig.directory)
			return False
		except OSError:
			error("Moira directory is not found or not set")
		return True

	#-----------------------------------------------------------------------
	@staticmethod
	def documentDir():
		return os.path.join(MoiraConfig.directory,"doc")

	#-----------------------------------------------------------------------
	# Return a command, default if cmd=empty
	#-----------------------------------------------------------------------
	@staticmethod
	def command(prg):
		if MoiraConfig.checkDir(): return prg
		if not prg:
			cmd = os.path.join(MoiraConfig.directory, prg)
			if os.path.isfile(cmd):
				return cmd
			return Config.command(prg)
		else:
			if prg[0] == os.path.sep:
				return prg
			else:
				return os.path.join(MoiraConfig.directory, prg)

	#-----------------------------------------------------------------------
	# Load from config
	#-----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		section = MoiraConfig.SECTION
		MoiraConfig.directory = config.getStr(section, "directory", MoiraConfig.directory)
		MoiraConfig.var       = config.getStr(section, "var",       MoiraConfig.var)
		MoiraConfig.program   = config.getStr(section, "program",   MoiraConfig.program)

	#-----------------------------------------------------------------------
	# Save to config
	#-----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass
