# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   17-Oct-2018

import re
import Evaluation
from deck import CardInfo, Card, CSGCard

#===============================================================================
# Global dictionary for penelope, containing all units constants
#===============================================================================
class PenelopeGlobalDict(Evaluation.GlobalDict):
	def __init__(self):
		super().__init__()
		self.redefineDistance_cm(1)	# in cm
		self.redefineEnergy_GeV(1e9)	# in eV
		self.redefineTime_s(1)		# in s

#===============================================================================
# Penelope CardInfo class
#===============================================================================
class PenelopeCardInfo(CardInfo):
	# ----------------------------------------------------------------------
	# Post initialization after loading other inputs init
	# ----------------------------------------------------------------------
	@staticmethod
	def post_init():
		section = "penelope"
		CardInfo.materials[section] = "MATERIAL"
		CardInfo.volumes[section]   = "BODY"

#===============================================================================
# PenelopeCard base class
#===============================================================================
class PenelopeCard(Card, CSGCard):
	"""Penelope input card"""
	MODE       = "penelope"
	COMMENT    = "       "
	DISABLE    = COMMENT

	cardInfo   = PenelopeCardInfo	# Pointer to Card Info
	globalDict = PenelopeGlobalDict()

	PAT_INLINE_CMT = re.compile(r"^\s*(\[.*\])\s*", re.M)

	# ======================================================================
	# String functions
	# ======================================================================
	# Return a formatted comment string
	# ----------------------------------------------------------------------
	def commentStr(self):
		"""return a comment string"""
		lines = []
		first_inline = True
		for line in self._comment.split("\n"):
			m = PenelopeCard.PAT_INLINE_CMT.match(line)
			if m and first_inline:
				first_inline = False
				continue
			line = line.rstrip()
			if line:
				lines.append(f"{self.COMMENT}{line}")
			else:
				lines.append(f"{self.COMMENT}.")
		if lines and lines[-1]: lines.append("")	# to append a last new line
		return "\n".join(lines)

	#-----------------------------------------------------------------------
	# return string with only the cards
	# ----------------------------------------------------------------------
	def _toStr(self, prefix=""):
		# find inline comment and append it
		for cmt in self.comment().splitlines():
			m = PenelopeCard.PAT_INLINE_CMT.match(cmt)
			if m:
				inline_cmt = m.group(1)
				break
		else:
			inline_cmt = None
		line = "%s%-7s%s"%(prefix, self.tag, " ".join(map(str,self.whats())))
		if inline_cmt:
			line += " "*(max(1,70-len(inline_cmt)-len(line)))
			line += inline_cmt
		return line.rstrip()
