# -*- coding: utf-8 -*-
#
# Copyright European Organization for Nuclear Research (CERN)
# All rights reserved
#
# Please look at the supplied documentation for the user's
# license
#
# Author: Vasilis.Vlachoudis@cern.ch
# Date:   23-Nov-2018

import os
import re

from log import error
from deck import Config

#===============================================================================
# A container class to keep the programs
# It should contain only static and class methods
#===============================================================================
class PenelopeConfig(Config):
	SECTION   = "Penelope"	# Configuration section name

	DEFAULT_EXE = "penelope"

	CYCLE     = "penelope_"	# prefix of temporary cycle directory

	directory = ""		# Executable installation directory
	var       = ""		# environment variable
	program   = DEFAULT_EXE	# Main application program
	version   = "Unknown"
	released  = ""

	CYCLEDIR_PAT  = re.compile(r"^%s(\d+)$"%(CYCLE))
	CYCLEDIR2_PAT = re.compile(r"^(.*/?)%s(\d+)$"%(CYCLE))

	#-----------------------------------------------------------------------
	@staticmethod
	def checkDir():
		try:
			os.stat(PenelopeConfig.directory)
			return False
		except OSError:
			error("Penelope directory is not found or not set")
		return True

	#-----------------------------------------------------------------------
	@staticmethod
	def documentDir():
		return os.path.join(PenelopeConfig.directory,"doc")

	#-----------------------------------------------------------------------
	# Return a command, default if cmd=empty
	#-----------------------------------------------------------------------
	@staticmethod
	def command(prg):
		if PenelopeConfig.checkDir(): return prg
		if not prg:
			cmd = os.path.join(PenelopeConfig.directory, prg)
			# Try fluka directory
			try:
				os.stat(cmd)
				return cmd
			except:
				pass
			return Config.command(prg)
		else:
			if prg[0] == os.path.sep:
				return prg
			else:
				return os.path.join(PenelopeConfig.directory, prg)

	#-----------------------------------------------------------------------
	# Load from config
	#-----------------------------------------------------------------------
	@staticmethod
	def loadConfig(config):
		section = PenelopeConfig.SECTION
		PenelopeConfig.directory = config.getStr(section, "directory", PenelopeConfig.directory)
		PenelopeConfig.var       = config.getStr(section, "var",       PenelopeConfig.var)
		PenelopeConfig.program   = config.getStr(section, "program",   PenelopeConfig.program)

	#-----------------------------------------------------------------------
	# Save to config
	#-----------------------------------------------------------------------
	@staticmethod
	def saveConfig(config):
		pass
